# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)
# ruff: noqa: E402, F401

import importlib.metadata

from . import (
    beam_center_finder,
    common,
    conversions,
    i_of_q,
    io,
    masking,
    normalization,
)
from .beam_center_finder import beam_center_from_center_of_mass
from .direct_beam import direct_beam
from .types import BackgroundSubtractedIofQ, IofQ, ReturnEvents, SampleRun
from .workflow import (
    with_background_runs,
    with_banks,
    with_pixel_mask_filenames,
    with_sample_runs,
)

try:
    __version__ = importlib.metadata.version("esssans")
except importlib.metadata.PackageNotFoundError:
    __version__ = "0.0.0"

providers = (
    *conversions.providers,
    *i_of_q.providers,
    *masking.providers,
    *normalization.providers,
    common.beam_center_to_detector_position_offset,
)
"""
List of providers for setting up a Sciline pipeline.

This provides a default workflow, including a beam-center estimation based on a
center-of-mass approach. Providers for loadings files are not included. Combine with
the providers for a specific instrument, such as :py:data:`esssans.sans2d.providers`
to setup a complete workflow.
"""

del importlib

__all__ = [
    'BackgroundSubtractedIofQ',
    'IofQ',
    'SampleRun',
    'ReturnEvents',
    'beam_center_finder',
    'beam_center_from_center_of_mass',
    'common',
    'conversions',
    'direct_beam',
    'i_of_q',
    'io',
    'masking',
    'normalization',
    'providers',
    'with_background_runs',
    'with_banks',
    'with_pixel_mask_filenames',
    'with_sample_runs',
]



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
"""
Default parameters, providers and utility functions for the loki workflow.
"""

import sciline
import scipp as sc

from ess.reduce.nexus.generic_workflow import GenericNeXusWorkflow
from ess.sans import providers as sans_providers
from ess.sans.io import read_xml_detector_masking

from ..sans.types import (
    CorrectForGravity,
    DetectorBankSizes,
    DetectorData,
    DetectorPixelShape,
    DimsToKeep,
    Incident,
    LabFrameTransform,
    MonitorData,
    MonitorType,
    NeXusDetector,
    NeXusMonitorName,
    NonBackgroundWavelengthRange,
    PixelShapePath,
    RunType,
    ScatteringRunType,
    TofData,
    TofMonitor,
    TransformationPath,
    Transmission,
    WavelengthBands,
    WavelengthMask,
)

DETECTOR_BANK_SIZES = {
    'larmor_detector': {'layer': 4, 'tube': 32, 'straw': 7, 'pixel': 512}
}


def default_parameters() -> dict:
    return {
        CorrectForGravity: False,
        DetectorBankSizes: DETECTOR_BANK_SIZES,
        DimsToKeep: (),
        NeXusMonitorName[Incident]: 'monitor_1',
        NeXusMonitorName[Transmission]: 'monitor_2',
        TransformationPath: 'transform',
        PixelShapePath: 'pixel_shape',
        NonBackgroundWavelengthRange: None,
        WavelengthMask: None,
        WavelengthBands: None,
    }


def _convert_to_tof(da: sc.DataArray) -> sc.DataArray:
    da.bins.coords['tof'] = da.bins.coords.pop('event_time_offset')
    if 'event_time_zero' in da.dims:
        da = da.bins.concat('event_time_zero')
    return da


def data_to_tof(
    da: DetectorData[ScatteringRunType],
) -> TofData[ScatteringRunType]:
    return TofData[ScatteringRunType](_convert_to_tof(da))


def monitor_to_tof(
    da: MonitorData[RunType, MonitorType],
) -> TofMonitor[RunType, MonitorType]:
    return TofMonitor[RunType, MonitorType](_convert_to_tof(da))


def detector_pixel_shape(
    detector: NeXusDetector[ScatteringRunType],
    pixel_shape_path: PixelShapePath,
) -> DetectorPixelShape[ScatteringRunType]:
    return DetectorPixelShape[ScatteringRunType](detector[pixel_shape_path])


def detector_lab_frame_transform(
    detector: NeXusDetector[ScatteringRunType],
    transform_path: TransformationPath,
) -> LabFrameTransform[ScatteringRunType]:
    return LabFrameTransform[ScatteringRunType](detector[transform_path])


loki_providers = (
    detector_pixel_shape,
    detector_lab_frame_transform,
    data_to_tof,
    monitor_to_tof,
)


def LokiAtLarmorWorkflow() -> sciline.Pipeline:
    """
    Workflow with default parameters for Loki test at Larmor.

    This version of the Loki workflow:

    - Uses ISIS XML files to define masks.
    - Sets a dummy sample position [0,0,0] since files do not contain this information.

    Returns
    -------
    :
        Loki workflow as a sciline.Pipeline
    """
    workflow = GenericNeXusWorkflow()
    for provider in sans_providers + loki_providers:
        workflow.insert(provider)
    for key, param in default_parameters().items():
        workflow[key] = param
    workflow.insert(read_xml_detector_masking)
    return workflow



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)
import sciline as sl

from ess.loki.workflow import LoKiMonitorWorkflow


def test_loki_monitor_workflow() -> None:
    # Test building the sciline pipeline.
    workflow = LoKiMonitorWorkflow()
    assert isinstance(workflow.pipeline, sl.Pipeline)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)
"""
File loading functions for ISIS data, NOT using Mantid.
"""

from typing import NewType

import sciline
import scipp as sc

from ess.sans.types import (
    BackgroundRun,
    DirectBeam,
    DirectBeamFilename,
    Filename,
    RunType,
    SampleRun,
    TransmissionRun,
)

CalibrationFilename = NewType('CalibrationFilename', str)


class LoadedFileContents(sciline.Scope[RunType, sc.DataGroup], sc.DataGroup):
    """Contents of a loaded file."""


def load_tutorial_run(filename: Filename[RunType]) -> LoadedFileContents[RunType]:
    return LoadedFileContents[RunType](sc.io.load_hdf5(filename))


def load_tutorial_direct_beam(filename: DirectBeamFilename) -> DirectBeam:
    return DirectBeam(sc.io.load_hdf5(filename))


def transmission_from_sample_run(
    data: LoadedFileContents[SampleRun],
) -> LoadedFileContents[TransmissionRun[SampleRun]]:
    """
    Use transmission from a sample run, instead of dedicated run.
    """
    return LoadedFileContents[TransmissionRun[SampleRun]](data)


def transmission_from_background_run(
    data: LoadedFileContents[BackgroundRun],
) -> LoadedFileContents[TransmissionRun[BackgroundRun]]:
    """
    Use transmission from a background run, instead of dedicated run.
    """
    return LoadedFileContents[TransmissionRun[BackgroundRun]](data)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

import scipp as sc
import scippnexus as snx
from scippnexus.application_definitions import nxcansas

from .types import (
    BackgroundSubtractedIofQ,
    MaskedDetectorIDs,
    OutFilename,
    PixelMaskFilename,
    RunNumber,
    RunTitle,
)


def save_background_subtracted_iofq(
    *,
    iofq: BackgroundSubtractedIofQ,
    out_filename: OutFilename,
    run_number: RunNumber,
    run_title: RunTitle,
) -> None:
    """Save background-subtracted I(Q) histogram as an NXcanSAS file."""
    if iofq.bins is None:
        da = iofq.copy(deep=False)
    else:
        da = iofq.hist()
    if da.coords.is_edges('Q'):
        da.coords['Q'] = sc.midpoints(da.coords['Q'])
    with snx.File(out_filename, 'w') as f:
        f['sasentry'] = nxcansas.SASentry(title=run_title, run=run_number)
        f['sasentry']['sasdata'] = nxcansas.SASdata(da, Q_variances='resolutions')


def read_xml_detector_masking(filename: PixelMaskFilename) -> MaskedDetectorIDs:
    """Read a pixel mask from an ISIS XML file.

    The format is as follows, where the detids are inclusive ranges of detector IDs:

    .. code-block:: xml

        <?xml version="1.0"?>
        <detector-masking>
            <group>
                <detids>1400203-1400218,1401199,1402190-1402223</detids>
            </group>
        </detector-masking>

    Parameters
    ----------
    filename:
        Path to the XML file.
    """
    import xml.etree.ElementTree as ET  # nosec

    tree = ET.parse(filename)  # noqa: S314
    root = tree.getroot()

    masked_detids = []
    for group in root.findall('group'):
        for detids in group.findall('detids'):
            for detid in detids.text.split(','):
                detid = detid.strip()
                if '-' in detid:
                    start, stop = detid.split('-')
                    masked_detids += list(range(int(start), int(stop) + 1))
                else:
                    masked_detids.append(int(detid))

    return MaskedDetectorIDs(
        sc.array(dims=['detector_id'], values=masked_detids, unit=None, dtype='int32')
    )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)

import importlib.metadata

from . import general, io, sans2d, zoom
from .general import default_parameters, SampleOffset, MonitorOffset, DetectorBankOffset
from .io import CalibrationFilename
from .visualization import plot_flat_detector_xy

try:
    __version__ = importlib.metadata.version(__package__ or __name__)
except importlib.metadata.PackageNotFoundError:
    __version__ = "0.0.0"

providers = general.providers

del importlib

__all__ = [
    'CalibrationFilename',
    'DetectorBankOffset',
    'MonitorOffset',
    'SampleOffset',
    'io',
    'providers',
    'plot_flat_detector_xy',
    'sans2d',
    'default_parameters',
    'zoom',
]



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
"""
Masking functions for the loki workflow.
"""

import numpy as np
import scipp as sc

from .types import (
    CalibratedDetector,
    DetectorIDs,
    DetectorMasks,
    MaskedData,
    MaskedDetectorIDs,
    PixelMaskFilename,
    SampleRun,
    ScatteringRunType,
    TofData,
)


def get_detector_ids_from_detector(data: CalibratedDetector[SampleRun]) -> DetectorIDs:
    """Extract detector IDs from a detector."""
    return DetectorIDs(
        data.coords[
            'detector_number' if 'detector_number' in data.coords else 'detector_id'
        ]
    )


def to_detector_mask(
    ids: DetectorIDs,
    name: PixelMaskFilename,
    masked_ids: MaskedDetectorIDs,
) -> DetectorMasks:
    """Create a detector mask from a list of masked detector IDs.
    The masks are based on detector IDs stored in XML files.

    Parameters
    ----------
    data:
        Raw data with configured component positions.
    masked_ids:
        Detector IDs to mask.
    """
    mask = sc.zeros(sizes=ids.sizes, dtype='bool')
    mask.values[np.isin(ids.values, masked_ids.values)] = True
    return DetectorMasks({name: mask})


def apply_pixel_masks(
    data: TofData[ScatteringRunType],
    masks: DetectorMasks,
) -> MaskedData[ScatteringRunType]:
    """Apply pixel-specific masks to raw data.

    Parameters
    ----------
    data:
        Raw data with configured component positions.
    masks:
        A series of masks.
    """
    return MaskedData[ScatteringRunType](data.assign_masks(masks))


providers = (
    get_detector_ids_from_detector,
    to_detector_mask,
    apply_pixel_masks,
)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)


class Registry:
    def __init__(self, instrument: str, files: dict[str, str], version: str):
        import pooch

        self._registry = pooch.create(
            path=pooch.os_cache(f'ess/{instrument}'),
            env=f'ESS_{instrument.upper()}_DATA_DIR',
            base_url=f'https://public.esss.dk/groups/scipp/ess/{instrument}/'
            + '{version}/',
            version=version,
            registry=files,
        )

    def __contains__(self, key):
        return key in self._registry.registry

    def get_path(self, name: str, unzip: bool = False) -> str:
        """
        Get the path to a file in the registry.

        Parameters
        ----------
        name:
            Name of the file to get the path for.
        unzip:
            If `True`, unzip the file before returning the path.
        """
        import pooch

        return self._registry.fetch(name, processor=pooch.Unzip() if unzip else None)


__all__ = ['Registry']



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

import importlib.metadata

from . import general
from .general import LokiAtLarmorWorkflow, default_parameters

try:
    __version__ = importlib.metadata.version(__package__ or __name__)
except importlib.metadata.PackageNotFoundError:
    __version__ = "0.0.0"

del importlib

__all__ = [
    'general',
    'default_parameters',
    'LokiAtLarmorWorkflow',
]



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

import pytest
import sciline
import scipp as sc
import scipp.testing
import scippnexus as snx
from scippnexus.application_definitions import nxcansas

from ess.sans.io import save_background_subtracted_iofq
from ess.sans.types import BackgroundSubtractedIofQ, OutFilename, RunNumber, RunTitle


@pytest.mark.parametrize('use_edges', [True, False])
def test_save_background_subtracted_iofq(use_edges, tmp_path):
    def background_subtracted_iofq() -> BackgroundSubtractedIofQ:
        i = sc.arange('Q', 0.0, 400.0)
        i.variances = i.values / 10
        return BackgroundSubtractedIofQ(
            sc.DataArray(
                i,
                coords={
                    'Q': sc.arange('Q', len(i) + int(use_edges), unit='1/angstrom')
                },
            )
        )

    def run_number() -> RunNumber:
        return RunNumber(7419)

    def run_title() -> RunTitle:
        return RunTitle('Test-title')

    out_filename = tmp_path / 'test.nxs'

    providers = (background_subtracted_iofq, run_number, run_title)
    params = {OutFilename: str(out_filename)}
    pipeline = sciline.Pipeline(providers=providers, params=params)
    pipeline.bind_and_call(save_background_subtracted_iofq)

    with snx.File(out_filename, 'r', definitions=nxcansas.definitions) as f:
        loaded = f[()]
    entry = loaded['sasentry']
    assert entry['title'] == 'Test-title'
    assert entry['run'] == 7419

    expected_data = background_subtracted_iofq()
    if use_edges:
        expected_data.coords['Q'] = sc.midpoints(expected_data.coords['Q'])
    sc.testing.assert_identical(entry['sasdata'].coords['Q'], expected_data.coords['Q'])
    # The conversion to stddevs in NeXus loses precision.
    assert sc.allclose(entry['sasdata'].data, expected_data.data)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
"""
Providers for the ISIS instruments.
"""

from typing import NewType

import sciline
import scipp as sc

from ess.sans.types import (
    BeamCenter,
    CalibratedDetector,
    CalibratedMonitor,
    CorrectForGravity,
    DetectorData,
    DetectorIDs,
    DetectorPixelShape,
    DetectorPositionOffset,
    DimsToKeep,
    Incident,
    LabFrameTransform,
    MonitorData,
    MonitorPositionOffset,
    MonitorType,
    NeXusDetector,
    NeXusMonitor,
    NeXusMonitorName,
    NonBackgroundWavelengthRange,
    RunNumber,
    RunTitle,
    RunType,
    SamplePosition,
    SampleRun,
    ScatteringRunType,
    SourcePosition,
    TofData,
    TofMonitor,
    Transmission,
    WavelengthBands,
    WavelengthMask,
)

from .io import LoadedFileContents
from .mantidio import Period


class MonitorOffset(sciline.Scope[MonitorType, sc.Variable], sc.Variable):
    """
    Offset for monitor position for all runs.
    """


DetectorBankOffset = NewType('DetectorBankOffset', sc.Variable)
SampleOffset = NewType('SampleOffset', sc.Variable)


def default_parameters() -> dict:
    return {
        CorrectForGravity: False,
        DimsToKeep: (),
        MonitorOffset[Incident]: MonitorOffset[Incident](
            sc.vector([0, 0, 0], unit='m')
        ),
        MonitorOffset[Transmission]: MonitorOffset[Transmission](
            sc.vector([0, 0, 0], unit='m')
        ),
        DetectorBankOffset: DetectorBankOffset(sc.vector([0, 0, 0], unit='m')),
        SampleOffset: SampleOffset(sc.vector([0, 0, 0], unit='m')),
        NonBackgroundWavelengthRange: None,
        WavelengthMask: None,
        WavelengthBands: None,
        Period: None,
    }


def to_detector_position_offset(
    global_offset: DetectorBankOffset, beam_center: BeamCenter
) -> DetectorPositionOffset[RunType]:
    return DetectorPositionOffset[RunType](global_offset - beam_center)


def to_monitor_position_offset(
    global_offset: MonitorOffset[MonitorType],
) -> MonitorPositionOffset[RunType, MonitorType]:
    return MonitorPositionOffset[RunType, MonitorType](global_offset)


def get_source_position(dg: LoadedFileContents[RunType]) -> SourcePosition[RunType]:
    """Get source position from raw data."""
    return SourcePosition[RunType](dg['data'].coords['source_position'])


def get_sample_position(
    dg: LoadedFileContents[RunType], offset: SampleOffset
) -> SamplePosition[RunType]:
    """Get sample position from raw data and apply user offset."""
    return SamplePosition[RunType](
        dg['data'].coords['sample_position'] + offset.to(unit='m')
    )


def get_detector_data(dg: LoadedFileContents[RunType]) -> NeXusDetector[RunType]:
    """Get detector data and apply user offsets to raw data.

    Parameters
    ----------
    dg:
        Data loaded with Mantid and converted to Scipp.
    """
    # The generic NeXus workflow will try to extract 'data' from this, which is exactly
    # what we also have in the Mantid data. We use the generic workflow since it also
    # applies offsets, etc.
    return NeXusDetector[RunType](dg)


def get_monitor_data(
    dg: LoadedFileContents[RunType], nexus_name: NeXusMonitorName[MonitorType]
) -> NeXusMonitor[RunType, MonitorType]:
    # The generic NeXus workflow will try to extract 'data' from this, which is exactly
    # what we also have in the Mantid data. We use the generic workflow since it also
    # applies offsets, etc.
    monitor = dg['monitors'][nexus_name]['data']
    return NeXusMonitor[RunType, MonitorType](
        sc.DataGroup(data=monitor, position=monitor.coords['position'])
    )


def dummy_assemble_detector_data(
    detector: CalibratedDetector[RunType],
) -> DetectorData[RunType]:
    """Dummy assembly of detector data, detector already contains neutron data."""
    return DetectorData[RunType](detector)


def dummy_assemble_monitor_data(
    monitor: CalibratedMonitor[RunType, MonitorType],
) -> MonitorData[RunType, MonitorType]:
    """Dummy assembly of monitor data, monitor already contains neutron data."""
    return MonitorData[RunType, MonitorType](monitor)


def data_to_tof(
    da: DetectorData[ScatteringRunType],
) -> TofData[ScatteringRunType]:
    """Dummy conversion of data to time-of-flight data.
    The data already has a time-of-flight coordinate."""
    return TofData[ScatteringRunType](da)


def monitor_to_tof(
    da: MonitorData[RunType, MonitorType],
) -> TofMonitor[RunType, MonitorType]:
    """Dummy conversion of monitor data to time-of-flight data.
    The monitor data already has a time-of-flight coordinate."""
    return TofMonitor[RunType, MonitorType](da)


def run_number(dg: LoadedFileContents[SampleRun]) -> RunNumber:
    """Get the run number from the raw sample data."""
    return RunNumber(int(dg['run_number']))


def run_title(dg: LoadedFileContents[SampleRun]) -> RunTitle:
    """Get the run title from the raw sample data."""
    return RunTitle(dg['run_title'].value)


def helium3_tube_detector_pixel_shape() -> DetectorPixelShape[ScatteringRunType]:
    # Pixel radius and length
    # found here:
    # https://github.com/mantidproject/mantid/blob/main/instrument/SANS2D_Definition_Tubes.xml
    R = 0.00405
    L = 0.002033984375
    pixel_shape = sc.DataGroup(
        {
            'vertices': sc.vectors(
                dims=['vertex'],
                values=[
                    # Coordinates in pixel-local coordinate system
                    # Bottom face center
                    [0, 0, 0],
                    # Bottom face edge
                    [R, 0, 0],
                    # Top face center
                    [0, L, 0],
                ],
                unit='m',
            ),
            'nexus_class': 'NXcylindrical_geometry',
        }
    )
    return pixel_shape


def lab_frame_transform() -> LabFrameTransform[ScatteringRunType]:
    # Rotate +y to -x
    return sc.spatial.rotation(value=[0, 0, 1 / 2**0.5, 1 / 2**0.5])


def get_detector_ids_from_sample_run(data: TofData[SampleRun]) -> DetectorIDs:
    """Extract detector IDs from sample run.

    This overrides the function in the masking module which gets the detector IDs from
    the detector before loading event data. In this ISIS case files are loaded using
    Mantid which does not load event separately, so we get IDs from the data.
    """
    return DetectorIDs(
        data.coords[
            'detector_number' if 'detector_number' in data.coords else 'detector_id'
        ]
    )


providers = (
    dummy_assemble_detector_data,
    dummy_assemble_monitor_data,
    to_detector_position_offset,
    to_monitor_position_offset,
    get_source_position,
    get_sample_position,
    get_detector_data,
    get_detector_ids_from_sample_run,
    get_monitor_data,
    data_to_tof,
    monitor_to_tof,
    run_number,
    run_title,
    lab_frame_transform,
    helium3_tube_detector_pixel_shape,
)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
from collections.abc import Hashable, Iterable

import pandas as pd
import sciline
import scipp as sc

from .types import (
    BackgroundRun,
    CleanSummedQ,
    Denominator,
    DetectorMasks,
    Filename,
    NeXusDetectorName,
    Numerator,
    PixelMaskFilename,
    SampleRun,
)


def _merge(*dicts: dict) -> dict:
    return {key: value for d in dicts for key, value in d.items()}


def merge_contributions(*data: sc.DataArray) -> sc.DataArray:
    if len(data) == 1:
        return data[0]
    reducer = sc.reduce(data)
    return reducer.bins.concat() if data[0].bins is not None else reducer.sum()


def with_pixel_mask_filenames(
    workflow: sciline.Pipeline, masks: Iterable[str]
) -> sciline.Pipeline:
    """
    Return modified workflow with pixel mask filenames set.

    Parameters
    ----------
    workflow:
        Workflow to modify.
    masks:
        List or tuple of pixel mask filenames to set.
    """
    workflow = workflow.copy()
    workflow[DetectorMasks] = (
        workflow[DetectorMasks]
        .map(pd.DataFrame({PixelMaskFilename: masks}).rename_axis('mask'))
        .reduce(index='mask', func=_merge)
    )
    return workflow


def with_banks(
    workflow: sciline.Pipeline,
    banks: Iterable[str],
    index: Iterable[Hashable] | None = None,
) -> sciline.Pipeline:
    """
    Return modified workflow with bank names set.

    Since banks typically have different Q-resolution the I(Q) of banks are not merged.
    That is, the resulting workflow will have separate outputs for each bank. Use
    :py:func:`sciline.compute_mapped` to compute results for all banks.

    Parameters
    ----------
    workflow:
        Workflow to modify.
    banks:
        List or tuple of bank names to set.
    index:
        Index to use for the DataFrame. If not provided, the bank names are used.
    """
    index = index or banks
    return workflow.map(
        pd.DataFrame({NeXusDetectorName: banks}, index=index).rename_axis('bank')
    )


def _set_runs(
    pipeline: sciline.Pipeline, runs: Iterable[str], key: Hashable, axis_name: str
) -> sciline.Pipeline:
    pipeline = pipeline.copy()
    runs = pd.DataFrame({Filename[key]: runs}).rename_axis(axis_name)
    for part in (Numerator, Denominator):
        pipeline[CleanSummedQ[key, part]] = (
            pipeline[CleanSummedQ[key, part]]
            .map(runs)
            .reduce(index=axis_name, func=merge_contributions)
        )
    return pipeline


def with_sample_runs(
    workflow: sciline.Pipeline, runs: Iterable[str]
) -> sciline.Pipeline:
    """
    Return modified workflow with sample run filenames set.

    Parameters
    ----------
    workflow:
        Workflow to modify.
    runs:
        List or tuple of sample run filenames to set.
    """
    return _set_runs(workflow, runs, SampleRun, 'sample_run')


def with_background_runs(
    workflow: sciline.Pipeline, runs: Iterable[str]
) -> sciline.Pipeline:
    """
    Return modified workflow with background run filenames set.

    Parameters
    ----------
    workflow:
        Workflow to modify.
    runs:
        List or tuple of background run filenames to set.
    """
    return _set_runs(workflow, runs, BackgroundRun, 'background_run')



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
import sciline
import scipp as sc

from ess import loki
from ess.sans.types import (
    BackgroundRun,
    CorrectForGravity,
    DetectorMasks,
    DirectBeam,
    EmptyBeamRun,
    Filename,
    NeXusDetectorName,
    QBins,
    QxBins,
    QyBins,
    ReturnEvents,
    SampleRun,
    TransmissionRun,
    UncertaintyBroadcastMode,
    WavelengthBins,
)


def make_workflow(no_masks: bool = True) -> sciline.Pipeline:
    wf = loki.LokiAtLarmorWorkflow()

    wf[NeXusDetectorName] = 'larmor_detector'
    wf[Filename[SampleRun]] = loki.data.loki_tutorial_sample_run_60339()
    wf[Filename[BackgroundRun]] = loki.data.loki_tutorial_background_run_60393()
    wf[Filename[TransmissionRun[SampleRun]]] = (
        loki.data.loki_tutorial_sample_transmission_run()
    )
    wf[Filename[TransmissionRun[BackgroundRun]]] = loki.data.loki_tutorial_run_60392()
    wf[Filename[EmptyBeamRun]] = loki.data.loki_tutorial_run_60392()

    wf[WavelengthBins] = sc.linspace(
        'wavelength', start=1.0, stop=13.0, num=51, unit='angstrom'
    )
    wf[CorrectForGravity] = True
    wf[UncertaintyBroadcastMode] = UncertaintyBroadcastMode.upper_bound
    wf[ReturnEvents] = False

    wf[QxBins] = sc.linspace('Qx', start=-0.3, stop=0.3, num=91, unit='1/angstrom')
    wf[QyBins] = sc.linspace('Qy', start=-0.2, stop=0.3, num=78, unit='1/angstrom')
    wf[QBins] = sc.linspace('Q', start=0.01, stop=0.3, num=101, unit='1/angstrom')
    # We have no direct-beam file for Loki currently
    wf[DirectBeam] = None
    if no_masks:
        wf[DetectorMasks] = {}

    return wf



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
import scipp as sc

from ess import sans

merge_func = sans.workflow.merge_contributions


def test_merge_banks_or_runs_returns_first_value_if_length_1() -> None:
    data = sc.DataArray(sc.scalar(1.0))
    assert merge_func(data) is data


def test_merge_banks_or_runs_sums_dense_data() -> None:
    data = [
        sc.DataArray(data=sc.array(dims=['Q'], values=[1.0, 2.0])),
        sc.DataArray(data=sc.array(dims=['Q'], values=[3.0, 4.0])),
    ]
    assert sc.identical(
        merge_func(*data),
        sc.DataArray(
            data=sc.array(dims=['Q'], values=[4.0, 6.0]),
        ),
    )


def test_merge_banks_or_runs_concats_bins() -> None:
    events1 = sc.DataArray(data=sc.array(dims=['event'], values=[1.0, 2.0, 3.0]))
    events2 = sc.DataArray(data=sc.array(dims=['event'], values=[4.0, 5.0]))
    data = [
        sc.bins(
            begin=sc.array(dims=['Q'], values=[0, 1], unit=None),
            dim='event',
            data=events1,
        ),
        sc.bins(
            begin=sc.array(dims=['Q'], values=[0, 2], unit=None),
            dim='event',
            data=events2,
        ),
    ]
    expected_events = sc.DataArray(
        data=sc.array(dims=['event'], values=[1.0, 4.0, 5.0, 2.0, 3.0]),
    )
    expected = sc.bins(
        begin=sc.array(dims=['Q'], values=[0, 3], unit=None),
        dim='event',
        data=expected_events,
    )
    assert sc.identical(merge_func(*data), expected)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)
from ess.sans.data import Registry
from ess.sans.types import (
    BackgroundRun,
    DirectBeamFilename,
    EmptyBeamRun,
    Filename,
    PixelMaskFilename,
    SampleRun,
)

from .io import CalibrationFilename

_sans2d_registry = Registry(
    instrument='sans2d',
    files={
        # Direct beam file (efficiency of detectors as a function of wavelength)
        'DIRECT_SANS2D_REAR_34327_4m_8mm_16Feb16.dat.h5': 'md5:43f4188301d709aa49df0631d03a67cb',  # noqa: E501
        # Empty beam run (no sample and no sample holder/can)
        'SANS2D00063091.nxs.h5': 'md5:ec7f78d51a4abc643bbe1965b7a876b9',
        # Sample run (sample and sample holder/can)
        'SANS2D00063114.nxs.h5': 'md5:d0701afe88c09e6a714b31ecfbd79c0c',
        # Background run (no sample, sample holder/can only)
        'SANS2D00063159.nxs.h5': 'md5:8d740b29d8965c8d9ca4f20f1e68ec15',
        # Solid angles of the SANS2D detector pixels computed by Mantid (for tests)
        'SANS2D00063091.SolidAngle_from_mantid.h5': 'md5:d57b82db377cb1aea0beac7202713861',  # noqa: E501
    },
    version='1',
)


def sans2d_solid_angle_reference() -> str:
    """Solid angles of the SANS2D detector pixels computed by Mantid (for tests)"""
    return _sans2d_registry.get_path('SANS2D00063091.SolidAngle_from_mantid.h5')


_zoom_registry = Registry(
    instrument='zoom',
    files={
        # Sample run (sample and sample holder/can) with applied 192tubeCalibration_11-02-2019_r5_10lines.nxs  # noqa: E501
        'ZOOM00034786.nxs.h5.zip': 'md5:e1c53bf826dd87545df1b3629f424762',
        # Empty beam run (no sample and no sample holder/can) - Scipp-hdf5 format
        'ZOOM00034787.nxs.h5': 'md5:27e563d4e57621518658307acbbc3413',
        # Calibration file, Mantid processed NeXus
        '192tubeCalibration_11-02-2019_r5_10lines.nxs': 'md5:ca1e0e3c387903be445d0dfd0a784ed6',  # noqa: E501
        # Direct beam file (efficiency of detectors as a function of wavelength)
        'Direct_Zoom_4m_8mm_100522.txt.h5': 'md5:bbe813580676a9ad170934ffb7c99617',
        # Moderator file (used for computing Q-resolution)
        'ModeratorStdDev_TS2_SANS_LETexptl_07Aug2015.txt': 'md5:5fc389340d453b9095a5dfcc33608dae',  # noqa: E501
        # ISIS user file configuring the data reduction
        'USER_ZOOM_Cabral_4m_TJump_233G_8x8mm_Small_BEAMSTOP_v1_M5.toml': 'md5:4423ecb7d924c79711aba5b0a30a23e7',  # noqa: E501
        # 7 pixel mask files for the ZOOM00034786.nxs run
        'andru_test.xml': 'md5:c59e0c4a80640a387df7beca4857e66f',
        'left_beg_18_2.xml': 'md5:5b24a8954d4d8a291f59f5392cd61681',
        'right_beg_18_2.xml': 'md5:fae95a5056e5f5ba4996c8dff83ec109',
        'small_bs_232.xml': 'md5:6d67dea9208193c9f0753ffcbb50ed83',
        'small_BS_31032023.xml': 'md5:3c644e8c75105809ab521773f9c0c85b',
        'tube_1120_bottom.xml': 'md5:fe577bf73c16bf5ac909516fa67360e9',
        'tubes_beg_18_2.xml': 'md5:2debde8d82c383cc3d592ea000552300',
    },
    version='2',
)


def sans2d_tutorial_direct_beam() -> DirectBeamFilename:
    return DirectBeamFilename(
        _sans2d_registry.get_path('DIRECT_SANS2D_REAR_34327_4m_8mm_16Feb16.dat.h5')
    )


def sans2d_tutorial_sample_run() -> Filename[SampleRun]:
    return Filename[SampleRun](_sans2d_registry.get_path('SANS2D00063114.nxs.h5'))


def sans2d_tutorial_background_run() -> Filename[BackgroundRun]:
    return Filename[BackgroundRun](_sans2d_registry.get_path('SANS2D00063159.nxs.h5'))


def sans2d_tutorial_empty_beam_run() -> Filename[EmptyBeamRun]:
    return Filename[EmptyBeamRun](_sans2d_registry.get_path('SANS2D00063091.nxs.h5'))


def zoom_tutorial_direct_beam() -> DirectBeamFilename:
    return DirectBeamFilename(
        _zoom_registry.get_path('Direct_Zoom_4m_8mm_100522.txt.h5')
    )


def zoom_tutorial_calibration() -> Filename[CalibrationFilename]:
    return Filename[CalibrationFilename](
        _zoom_registry.get_path('192tubeCalibration_11-02-2019_r5_10lines.nxs')
    )


def zoom_tutorial_sample_run() -> Filename[SampleRun]:
    return Filename[SampleRun](
        _zoom_registry.get_path('ZOOM00034786.nxs.h5.zip', unzip=True)[0]
    )


def zoom_tutorial_empty_beam_run() -> Filename[EmptyBeamRun]:
    return Filename[EmptyBeamRun](_zoom_registry.get_path('ZOOM00034787.nxs.h5'))


def zoom_tutorial_mask_filenames() -> list[PixelMaskFilename]:
    return [
        PixelMaskFilename(_zoom_registry.get_path('andru_test.xml')),
        PixelMaskFilename(_zoom_registry.get_path('left_beg_18_2.xml')),
        PixelMaskFilename(_zoom_registry.get_path('right_beg_18_2.xml')),
        PixelMaskFilename(_zoom_registry.get_path('small_bs_232.xml')),
        PixelMaskFilename(_zoom_registry.get_path('small_BS_31032023.xml')),
        PixelMaskFilename(_zoom_registry.get_path('tube_1120_bottom.xml')),
        PixelMaskFilename(_zoom_registry.get_path('tubes_beg_18_2.xml')),
    ]
